import {
  AuthorizeSecurityGroupIngressCommand,
  CreateKeyPairCommand,
  CreateSecurityGroupCommand,
  DeleteKeyPairCommand,
  DeleteSecurityGroupCommand,
  DescribeInstancesCommand,
  DescribeSecurityGroupsCommand,
  EC2Client,
  RevokeSecurityGroupIngressCommand,
  RunInstancesCommand,
  TerminateInstancesCommand,
  _InstanceType,
} from '@aws-sdk/client-ec2'
import { AuxVMDetails, GPUSpec, VMSpec } from '../Driver'

type GPUDetail = {
  count: number
  model: NonNullable<GPUSpec['model']>
}

type InstanceDetail = {
  type: _InstanceType
  cpuCount: number
  ramGiB: number
  cpuArchitecture: NonNullable<VMSpec['cpu_architecture']>
  gpuDetail?: GPUDetail
}

const instanceDetails: InstanceDetail[] = [
  { type: 't3.nano', cpuCount: 2, ramGiB: 0.5, cpuArchitecture: 'x64' },
  { type: 't3.micro', cpuCount: 2, ramGiB: 1, cpuArchitecture: 'x64' },
  { type: 't3.small', cpuCount: 2, ramGiB: 2, cpuArchitecture: 'x64' },
  { type: 't3.medium', cpuCount: 2, ramGiB: 4, cpuArchitecture: 'x64' },
  { type: 't3.large', cpuCount: 2, ramGiB: 8, cpuArchitecture: 'x64' },
  { type: 't3.xlarge', cpuCount: 4, ramGiB: 16, cpuArchitecture: 'x64' },
  { type: 't3.2xlarge', cpuCount: 8, ramGiB: 32, cpuArchitecture: 'x64' },

  { type: 't4g.nano', cpuCount: 2, ramGiB: 0.5, cpuArchitecture: 'arm64' },
  { type: 't4g.micro', cpuCount: 2, ramGiB: 1, cpuArchitecture: 'arm64' },
  { type: 't4g.small', cpuCount: 2, ramGiB: 2, cpuArchitecture: 'arm64' },
  { type: 't4g.medium', cpuCount: 2, ramGiB: 4, cpuArchitecture: 'arm64' },
  { type: 't4g.large', cpuCount: 2, ramGiB: 8, cpuArchitecture: 'arm64' },
  { type: 't4g.xlarge', cpuCount: 4, ramGiB: 16, cpuArchitecture: 'arm64' },
  { type: 't4g.2xlarge', cpuCount: 8, ramGiB: 32, cpuArchitecture: 'arm64' },

  { type: 'g5.xlarge', cpuCount: 4, ramGiB: 16, cpuArchitecture: 'x64', gpuDetail: { count: 1, model: 'a10' } },
  { type: 'g5.2xlarge', cpuCount: 8, ramGiB: 32, cpuArchitecture: 'x64', gpuDetail: { count: 1, model: 'a10' } },
  { type: 'g5.4xlarge', cpuCount: 16, ramGiB: 64, cpuArchitecture: 'x64', gpuDetail: { count: 1, model: 'a10' } },
  { type: 'g5.8xlarge', cpuCount: 32, ramGiB: 128, cpuArchitecture: 'x64', gpuDetail: { count: 1, model: 'a10' } },
  {
    type: 'g5.16xlarge',
    cpuCount: 64,
    ramGiB: 256,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 1, model: 'a10' },
  },
  {
    type: 'g5.12xlarge',
    cpuCount: 48,
    ramGiB: 192,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 4, model: 'a10' },
  },
  {
    type: 'g5.24xlarge',
    cpuCount: 96,
    ramGiB: 384,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 4, model: 'a10' },
  },
  {
    type: 'g5.48xlarge',
    cpuCount: 192,
    ramGiB: 768,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 8, model: 'a10' },
  },

  {
    type: 'p4d.24xlarge',
    cpuCount: 96,
    ramGiB: 1152,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 8, model: 'a100' },
  },

  {
    type: 'p5.48xlarge',
    cpuCount: 192,
    ramGiB: 2048,
    cpuArchitecture: 'x64',
    gpuDetail: { count: 8, model: 'h100' },
  },
]

function isInRange(value: number, range: [number, number]) {
  return range[0] <= value && value <= range[1]
}

function doInstanceDetailsMatchGPUSpec(instanceDetails: InstanceDetail, vmSpec: VMSpec) {
  if (!vmSpec.gpu_spec) {
    return !instanceDetails.gpuDetail
  }

  if (!instanceDetails.gpuDetail) {
    return false
  }

  return (
    isInRange(instanceDetails.gpuDetail.count, vmSpec.gpu_spec.count_range) &&
    instanceDetails.gpuDetail.model === vmSpec.gpu_spec.model
  )
}

type AMIDetail = {
  id: string
  cpuArchitecture: NonNullable<VMSpec['cpu_architecture']>
  baseImageType: NonNullable<VMSpec['base_image_type']>
  // If unset, we assume that the AMI supports all instance types for the given CPU architecture.
  supportedInstanceTypes?: _InstanceType[]
  sshUsername: string
}

const amiDetails: AMIDetail[] = [
  { id: 'ami-037071a31045dfccc', cpuArchitecture: 'x64', baseImageType: 'debian-12', sshUsername: 'admin' },
  { id: 'ami-093939e1f4317142c', cpuArchitecture: 'arm64', baseImageType: 'debian-12', sshUsername: 'admin' },
  {
    id: 'ami-0ce85c8d26b5f523b',
    cpuArchitecture: 'x64',
    baseImageType: 'ubuntu-20.04-cuda',
    supportedInstanceTypes: [
      'g5.xlarge',
      'g5.2xlarge',
      'g5.4xlarge',
      'g5.8xlarge',
      'g5.16xlarge',
      'g5.12xlarge',
      'g5.24xlarge',
      'g5.48xlarge',
    ],
    sshUsername: 'ubuntu',
  },
  {
    id: 'ami-0e575809a4bbc7dd1',
    cpuArchitecture: 'x64',
    baseImageType: 'ubuntu-20.04-cuda',
    supportedInstanceTypes: ['p4d.24xlarge', 'p5.48xlarge'],
    sshUsername: 'ubuntu',
  },
]

const DEFAULT_BASE_IMAGE_TYPE = 'debian-12'

function getInstanceTypeAndImage(vmSpec: VMSpec) {
  const matchingInstanceDetails = instanceDetails.filter(instanceDetails => {
    return (
      isInRange(instanceDetails.cpuCount, vmSpec.cpu_count_range) &&
      isInRange(instanceDetails.ramGiB, vmSpec.ram_gib_range) &&
      (vmSpec.cpu_architecture ?? 'arm64') === instanceDetails.cpuArchitecture &&
      doInstanceDetailsMatchGPUSpec(instanceDetails, vmSpec)
    )
  })

  if (matchingInstanceDetails.length === 0) {
    throw new Error(`No instance type found for VMSpec: ${JSON.stringify(vmSpec)}`)
  }

  const instanceDetail = matchingInstanceDetails[0]

  const matchingAMIDetails = amiDetails.filter(amiDetail => {
    return (
      amiDetail.cpuArchitecture === instanceDetail.cpuArchitecture &&
      amiDetail.baseImageType === (vmSpec.base_image_type ?? DEFAULT_BASE_IMAGE_TYPE) &&
      (amiDetail.supportedInstanceTypes === undefined || amiDetail.supportedInstanceTypes.includes(instanceDetail.type))
    )
  })
  if (matchingAMIDetails.length === 0) {
    throw new Error(`No AMI found for VMSpec: ${JSON.stringify(vmSpec)}`)
  }

  const amiDetail = matchingAMIDetails[0]

  return { imageId: amiDetail.id, sshUsername: amiDetail.sshUsername, instanceType: instanceDetail.type }
}

function getSecurityGroupIngress(groupId: string) {
  return {
    GroupId: groupId,
    IpPermissions: [
      // Allow all IPv4 addresses to connect to the VM over SSH.
      {
        IpProtocol: 'tcp',
        FromPort: 22,
        ToPort: 22,
        IpRanges: [{ CidrIp: '0.0.0.0/0' }],
      },
    ],
  }
}

export async function createEC2Instance(taskEnvironmentIdentifier: string, vmSpec: VMSpec): Promise<AuxVMDetails> {
  const ec2Client = new EC2Client()

  const tags = [
    {
      Key: 'Project',
      Value: 'MP4',
    },
    {
      Key: 'Name',
      Value: taskEnvironmentIdentifier,
    },
    {
      Key: 'Purpose',
      Value: 'TaskEnvironment',
    },
  ]

  const createKeyPairCommand = new CreateKeyPairCommand({
    KeyName: taskEnvironmentIdentifier,
    TagSpecifications: [
      {
        ResourceType: 'key-pair',
        Tags: tags,
      },
    ],
  })

  const keyPair = await ec2Client.send(createKeyPairCommand)

  const createSecurityGroupCommand = new CreateSecurityGroupCommand({
    GroupName: taskEnvironmentIdentifier,
    Description: `For SSH access to task environment VM ${taskEnvironmentIdentifier}`,
    TagSpecifications: [
      {
        ResourceType: 'security-group',
        Tags: tags,
      },
    ],
  })
  const securityGroup = await ec2Client.send(createSecurityGroupCommand)

  const authorizeSecurityGroupIngressCommand = new AuthorizeSecurityGroupIngressCommand(
    getSecurityGroupIngress(securityGroup.GroupId!)
  )
  await ec2Client.send(authorizeSecurityGroupIngressCommand)

  const { imageId, sshUsername, instanceType } = getInstanceTypeAndImage(vmSpec)

  const runInstancesCommand = new RunInstancesCommand({
    TagSpecifications: [
      {
        ResourceType: 'instance',
        Tags: tags,
      },
    ],
    ImageId: imageId,
    InstanceType: instanceType,
    MinCount: 1,
    MaxCount: 1,
    KeyName: taskEnvironmentIdentifier,
    NetworkInterfaces: [
      {
        DeviceIndex: 0, // Primary network interface
        AssociatePublicIpAddress: true,
        Groups: [securityGroup.GroupId!],
        DeleteOnTermination: true,
      },
    ],
  })
  const instance = await ec2Client.send(runInstancesCommand)

  const describeInstancesCommand = new DescribeInstancesCommand({
    InstanceIds: [instance.Instances![0].InstanceId!],
  })
  let ipAddress = null
  let retries = 0

  while (ipAddress === null) {
    try {
      const instances = await ec2Client.send(describeInstancesCommand)
      ipAddress = instances.Reservations?.[0].Instances?.[0].PublicIpAddress ?? null
    } catch (e) {
      if (retries >= 30) {
        throw e
      }

      retries += 1

      // Wait three seconds between retries to ensure that we wait at least 90 seconds for the instances to start.
      await new Promise(resolve => setTimeout(resolve, 3_000))
    }
  }

  return {
    sshUsername,
    sshPrivateKey: keyPair.KeyMaterial!,
    ipAddress: ipAddress!,
  }
}

export async function destroyEC2Instance(taskEnvironmentIdentifier: string) {
  const ec2Client = new EC2Client()

  const describeInstancesCommand = new DescribeInstancesCommand({
    Filters: [
      {
        Name: 'tag:Name',
        Values: [taskEnvironmentIdentifier],
      },
    ],
  })
  const instances = await ec2Client.send(describeInstancesCommand)

  const instanceIds =
    instances.Reservations?.flatMap(
      reservation => reservation.Instances?.map(instance => instance.InstanceId!) ?? []
    ) ?? []
  if (instanceIds.length === 0) {
    return
  }

  const terminateInstancesCommand = new TerminateInstancesCommand({
    InstanceIds: instanceIds,
  })
  await ec2Client.send(terminateInstancesCommand)

  const describeInstancesCommandForInstanceState = new DescribeInstancesCommand({
    InstanceIds: instanceIds,
  })
  let retries = 0

  while (true) {
    const response = await ec2Client.send(describeInstancesCommandForInstanceState)
    const instances = response.Reservations?.flatMap(reservation => reservation.Instances ?? []) ?? []
    if (instances.every(instance => instance.State?.Name === 'terminated')) {
      break
    }

    if (retries >= 30) {
      throw new Error('Failed to terminate instances')
    }

    retries += 1

    // Wait three seconds between retries to ensure that we wait at least 90 seconds for the instances to terminate.
    await new Promise(resolve => setTimeout(resolve, 3_000))
  }

  const describeSecurityGroupsCommand = new DescribeSecurityGroupsCommand({
    GroupNames: [taskEnvironmentIdentifier],
  })
  const securityGroupResponse = await ec2Client.send(describeSecurityGroupsCommand)
  const securityGroupId = securityGroupResponse.SecurityGroups![0].GroupId!

  const revokeSecurityGroupIngressCommand = new RevokeSecurityGroupIngressCommand(
    getSecurityGroupIngress(securityGroupId)
  )
  await ec2Client.send(revokeSecurityGroupIngressCommand)

  const deleteSecurityGroupCommand = new DeleteSecurityGroupCommand({
    GroupId: securityGroupId,
  })
  await ec2Client.send(deleteSecurityGroupCommand)

  const deleteKeyPairCommand = new DeleteKeyPairCommand({
    KeyName: taskEnvironmentIdentifier,
  })
  await ec2Client.send(deleteKeyPairCommand)
}
